/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.client;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

//SONAR-IGNORE-FILE

/**
 * Utility class that parses JSON string into Map<String,Object> representation, arrays are parsed as List<Object>
 * Based on JSON parser and formatter in tdp-mvc-util 2.1.x library
 */
final class JSONUtil
{
  static JSONValue parse(String json) throws KShieldParserException {
    if (json == null) throw new IllegalArgumentException("Parameter json is must not be null");
    
    final ParserContext ctx = new RootContext(json.toCharArray(), -1);
    
    parseJsonValue(ctx);
    
    consumeTrailingWhitespaces(ctx);
    
    return ctx.value();
  }

  private static void consumeTrailingWhitespaces(ParserContext ctx) throws KShieldParserException {
    while (ctx.hasNext()) {
      final char ch = ctx.next();
      
      if (!Character.isWhitespace(ch)) {
        throw new KShieldParserException("JSON Syntax Error: Unexpected character after JSON [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
      }
    }
  }
  
  private static void consumeWhitespaces(ParserContext ctx) {
    while (ctx.hasNext()) {
      final char ch = ctx.next();
      
      if (!Character.isWhitespace(ch)) {
        ctx.back();
        return;
      }
    }
  }
  
  private static void parseJsonValue(ParserContext ctx) throws KShieldParserException {
    consumeWhitespaces(ctx);
    
    while (ctx.hasNext()) {
      final char ch = ctx.next();
      
      switch (ch) {
        case '{':
          parseJsonBean(new HashContext(ctx));
          return;
          
        case '[':
          parseJsonArray(new ArrayContext(ctx));
          return;
          
        default:
          ctx.back();
          parseJsonString(new PlainContext(ctx));
          return;
      }
    }
  }
  
  private static void parseJsonString(PlainContext ctx) throws KShieldParserException {
    while (ctx.hasNext()) {
      final char ch = ctx.next();
      
      if (Character.isWhitespace(ch) && ctx.buf.length() > 0) {
        ctx.state = State.AFTER_VALUE;
        return;
      }
      
      switch (ch) {
        case ',':
          // end of string
          ctx.back();
          return;
          
        case '}':
          // and of hash
          ctx.back();
          return;
          
        case ']':
          // end of array
          ctx.back();
          return;
          
        case '\'':
          parseSingleQuotedValue(ctx.buf, ctx);
          return;
          
        case '"':
          parseDoubleQuotedValue(ctx.buf, ctx);
          return;
          
        default:
          if (Character.isLetterOrDigit(ch)) {
            // we allow only alphanumeric chars in unqoted values
            ctx.buf.append(ch);
            break;
          }
          
          throw new KShieldParserException("JSON Syntax Error: Illegal character in unquoted value [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
      }
    }
  }

  private static void parseJsonBean(HashContext ctx) throws KShieldParserException {
    final StringBuilder property = new StringBuilder();
    
    while (ctx.hasNext()) {
      consumeWhitespaces(ctx);
      
      final char ch = ctx.next();
      
      switch (ch) {
        case ',':
          if (ctx.valueState()) {
            // end of value => next is property
            ctx.state = State.PROPERTY;
            break;
          }
          
          throw new KShieldParserException("JSON Syntax Error: Unexpected comma character inside hash [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
          
        case '}':
          if (ctx.state == State.PROPERTY || ctx.valueState()) {
            // end of hash
            ctx.state = State.AFTER_VALUE;
            return;
          }
          
          throw new KShieldParserException("JSON Syntax Error: Unexpected end of nested hash value [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
          
       case ':':
          if (ctx.propertyState()) {
            parseNestedValue(property.toString(), ctx);
            break;
          }
          
        default:
          ctx.back();
          parseProperty(property, ctx);
          break;
      }
    }
  }
  
  private static void parseProperty(StringBuilder property, HashContext ctx) throws KShieldParserException {
    property.setLength(0);
    
    while (ctx.hasNext()) {
      final char ch = ctx.next();
      
      if (Character.isWhitespace(ch) && property.length() > 0) {
        ctx.state = State.AFTER_PROPERTY;
          
        // property finished
        return;
      }
      
      switch (ch) {
        case '}':
          if (property.length() == 0) {
            ctx.back();
            
            // end of hash before property start
            return;
          }
          
          throw new KShieldParserException("JSON Syntax Error: Unexpected character inside property [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
          
        case ':':
          ctx.back();
          ctx.state = State.AFTER_PROPERTY;
          // property finished
          return;
          
        case '\'':
          parseSingleQuotedProperty(property, ctx);
          break;
          
        case '"':
          parseDoubleQuotedProperty(property, ctx);
          break;
          
        default:
          if (Character.isJavaIdentifierStart(ch) || property.length() > 0 && Character.isJavaIdentifierPart(ch)) {
            property.append(ch);
            break;
          }
          
          throw new KShieldParserException("JSON Syntax Error: Unexpected character inside property [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
      }
    }
    
  }

  private static void parseSingleQuotedProperty(StringBuilder buf, HashContext ctx) {
    parseQuotedToken(buf, ctx, '\'');
    ctx.state = State.AFTER_PROPERTY;
  }
  
  private static void parseSingleQuotedValue(StringBuilder buf, AbstractContext ctx) {
    parseQuotedToken(buf, ctx, '\'');
    ctx.state = State.SINGLE_QUOTED_VALUE;
  }
  
  private static void parseDoubleQuotedProperty(StringBuilder buf, HashContext ctx) {
    parseQuotedToken(buf, ctx, '"');
    ctx.state = State.AFTER_PROPERTY;
  }
  
  private static void parseDoubleQuotedValue(StringBuilder buf, AbstractContext ctx) {
    parseQuotedToken(buf, ctx, '"');
    ctx.state = State.DOUBLE_QUOTED_VALUE;
  }
  
  private static void parseQuotedToken(StringBuilder buf, AbstractContext ctx, char quotChar) {
    buf.setLength(0);
    
    while (ctx.hasNext()) {
      final char ch = ctx.next();
      
      if (ctx.escape) {
        ctx.escape = false;
        buf.append(ch);
        continue;
      }
      
      switch (ch) {
        case '\\':
          ctx.escape = true;
          break;
          
        case '\'':
          if (quotChar == '\'') {
            // finished quoted token
            return;
          }
          
          buf.append(ch);
          break;
          
        case '"':
          if (quotChar == '"') {
            // finished quoted token
            return;
          }
          
          buf.append(ch);
          break;
          
        default:
          buf.append(ch);
          break;
      }
    }
  }

  private static void parseJsonArray(ArrayContext ctx) throws KShieldParserException {
    while (ctx.hasNext()) {
      consumeWhitespaces(ctx);
      
      if (ctx.array.list.isEmpty() && ctx.peek() == ']') {
        ctx.next();
        
        // end of empty array
        return;
      }
      
      parseNestedValue(ctx);
      consumeWhitespaces(ctx);
      
      final char ch = ctx.next();
      
      switch (ch) {
        case ',':
          // next value
          break;
          
        case ']':
          // end of array
          return;
          
        default:
          throw new KShieldParserException("JSON Syntax Error: Unexpected character in array value [offset="+ctx.position()+", ch="+ch+", ctx="+ctx+"]");
      }
    }
    
    
  }

  private static void parseNestedValue(String property, HashContext ctx) throws KShieldParserException {
    final NestedContext root = new NestedContext(ctx);
    parseJsonValue(root);
    ctx.finishValue(property, root);
  }
  
  private static void parseNestedValue(ArrayContext ctx) throws KShieldParserException {
    final NestedContext root = new NestedContext(ctx);
    parseJsonValue(root);
    ctx.finishValue(root);
  }

  static JSONValue handleNullValue(String value) {
    if (value == null || "null".equals(value)) {
      return JSONNull.NULL;
    }
    return new JSONString(value);
  }
  
  static String delimitJson(String value) {
    return "\"" + value + "\"";
  }
  
  public static void endObject(StringBuilder json) {
    final int endOffset = json.length() - 1;
    if (endOffset > 0 && json.charAt(endOffset) == ',') {
      json.deleteCharAt(endOffset);
    }
    
    json.append('}');
  }

  public static void startObject(StringBuilder json) {
    json.append('{');
  }

  public static void appendProperty(StringBuilder json, String key, String message) {
    JSONUtil.delimit(json, key);
    json.append(':');
    JSONUtil.delimit(json, message);
    json.append(',');
  }

  public static void delimit(StringBuilder json, String key) {
    json.append('"');
    json.append(key);
    json.append('"');
  }

  private static enum State
  {
    /**
     * Inside single quoted property identifier, SQUOTE => AFTER_PROPERTY | continue
     */
    SINGLE_QUOTED_PROPERTY,
    
    /**
     * Inside double quoted property identifier, DQUOTE => AFTER_PROPERTY | continue
     */
    DOUBLE_QUOTED_PROPERTY,
    
    /**
     * Inside single quoted value literal, SQUOTE => AFTER_VALUE | continue
     */
    SINGLE_QUOTED_VALUE,

    /**
     * Inside double quoted value literal, DQUOTE => AFTER_VALUE | continue
     */
    DOUBLE_QUOTED_VALUE,
    
    /**
     * At property identifier, COLON => VALUE | SQUOTE => SINGLE_QUOTED_PROPERTY | DQUOTE => DOUBLE_QUOTED_PROPERTY | ERROR
     */
    PROPERTY,
    
    /**
     * At value literal, COMMA (in array) => AFTER_VALUE | COMMA (in hash) => AFTER_PAIR | SQUOTE => SINGLE_QUOTED_VALUE | DQUOTE => DOUBLE_QUOTED_VALUE | ERROR
     */
    VALUE,
    
    /**
     * After value literal in array, ARRAY_END => END | COMMA => VALUE | ERROR
     */
    AFTER_VALUE,
    
    /**
     * After property identifier in hash, COLON => VALUE | ERROR
     */
    AFTER_PROPERTY,
    
    /**
     * After property/value pair in hash, HASH_END => END | COMMA => PROPERTY | ERROR
     */
    AFTER_PAIR
  }
  
  interface JSONValue
  {
    String asString();
    boolean asBoolean();
    int asInt();
    
    boolean isHash();
    JSONHash asHash();
    
    boolean isArray();
    JSONArray asArray();
    
    boolean isNull();
  }
  
  static class BaseJSONValue implements JSONValue
  {
    @Override
    public String asString() {
      return null;
    }

    @Override
    public boolean asBoolean() {
      return false;
    }

    @Override
    public int asInt() {
      return 0;
    }
    
    @Override
    public boolean isNull() {
      return false;
    }

    @Override
    public boolean isHash() {
      return false;
    }

    @Override
    public JSONHash asHash() {
      return JSONHash.EMPTY;
    }

    @Override
    public boolean isArray() {
      return false;
    }

    @Override
    public JSONArray asArray() {
      return JSONArray.EMPTY;
    }
  }
  
  static class JSONNull extends BaseJSONValue
  {
    static final JSONNull NULL = new JSONNull();

    @Override
    public boolean isNull() {
      return true;
    }
  }
  
  static class JSONString extends BaseJSONValue
  {
    static final JSONString EMPTY = new JSONString("");
    
    protected final String value;
    
    protected JSONString(String v) {
      value = v;
    }
    
    @Override
    public String toString() {
      return delimitJson(value);
    }

    @Override
    public String asString() {
      return value;
    }
    
    @Override
    public boolean asBoolean() {
      return Boolean.parseBoolean(value);
    }
    
    @Override
    public int asInt() {
      return Integer.parseInt(value, 10);
    }
  }
  
  static class JSONHash extends BaseJSONValue
  {
    static final JSONHash EMPTY = new JSONHash();
    
    protected final Map<String,JSONValue> map;
    
    protected JSONHash() {
      map = new HashMap<String,JSONValue>();
    }
    
    @Override
    public String asString() {
      return toString();
    }
    
    @Override
    public boolean isHash() {
      return true;
    }
    
    @Override
    public JSONHash asHash() {
      return this;
    }
    
    public Map<String,JSONValue> hash() {
      return Collections.unmodifiableMap(map);
    }
    
    public JSONValue get(String key) {
      if (map.containsKey(key)) {
        return map.get(key);
      }
      
      return JSONNull.NULL;
    }
    
    public int size() {
      return map.size();
    }
    
    @Override
    public String toString() {
      final StringBuilder buf = new StringBuilder();
      
      buf.append('{');
      
      if (!map.isEmpty()) {
        for (Map.Entry<String,JSONValue> entry : map.entrySet()) {
          buf.append(delimitJson(entry.getKey()));
          buf.append(':');
          buf.append(entry.getValue());
          buf.append(',');
        }
        
        buf.deleteCharAt(buf.length() - 1);
      }
      
      buf.append('}');
      
      return buf.toString();
    }
    
    protected void put(String key, JSONValue val) {
      map.put(key, val);
    }
  }
  
  static class JSONArray extends BaseJSONValue
  {
    static final JSONArray EMPTY = new JSONArray();
    
    protected final List<JSONValue> list;
    
    protected JSONArray() {
      list = new ArrayList<JSONValue>();
    }
    
    @Override
    public String asString() {
      return toString();
    }
    
    @Override
    public boolean isArray() {
      return true;
    }
    
    @Override
    public JSONArray asArray() {
      return this;
    }
    
    public List<JSONValue> values() {
      return Collections.unmodifiableList(list);
    }
    
    public JSONValue get(int index) {
      return list.get(index);
    }
    
    public int size() {
      return list.size();
    }
    
    @Override
    public String toString() {
      final StringBuilder buf = new StringBuilder();
      
      buf.append('[');
      
      if (!list.isEmpty()) {
        for (JSONValue value : list) {
          buf.append(value);
          buf.append(',');
        }
        
        buf.deleteCharAt(buf.length() - 1);
      }
      
      buf.append(']');
      
      return buf.toString();
    }
    
    protected void add(JSONValue val) {
      list.add(val);
    }
  }

  private interface ParserContext
  {
    JSONValue value();
    
    int position();
    
    char peek();
    
    char next();
    
    void back();
    
    boolean hasNext();

    void init(ParserContext ctx);
  }
  
  private static class RootContext implements ParserContext
  {
    final char[] chars;
    ParserContext valueContext;
    
    int pos;
    
    RootContext(char[] chars, int pos) {
      this.chars = chars;
      this.pos = pos;
    }
    
    @Override
    public void init(ParserContext ctx) {
      this.valueContext = ctx;
    }

    @Override
    public JSONValue value() {
      if (valueContext != null) {
        return valueContext.value();
      }
      
      return JSONString.EMPTY;
    }
    
    @Override
    public int position() {
      return pos;
    }
    
    @Override
    public char next() {
      return chars[++this.pos];
    }
    
    @Override
    public char peek() {
      return chars[this.pos + 1];
    }
    
    @Override
    public void back() {
      if (this.pos >= 0) {
        --this.pos;
      }
    }
    
    @Override
    public boolean hasNext() {
      return (pos+1) < chars.length;
    }
    
    @Override
    public String toString() {
      return "'" + String.valueOf(chars, pos, chars.length - pos) + "'";
    }
  }
  
  private abstract static class AbstractContext implements ParserContext
  {
    final ParserContext parent;
        
    State state = State.AFTER_VALUE;
    boolean escape = false;
    
    AbstractContext(ParserContext ctx) {
      parent = ctx;
      parent.init(this);
    }
    
    @Override
    public void init(ParserContext ctx) {
      // do nothing in base impl
    }
    
    @Override
    public int position() {
      return parent.position();
    }

    @Override
    public char peek() {
      return parent.peek();
    }
    
    @Override
    public char next() {
      return parent.next();
    }
    
    @Override
    public void back() {
      parent.back();
    }

    @Override
    public boolean hasNext() {
      return parent.hasNext();
    }
    
    public boolean quoted() {
      return QUOTED.contains(state);
    }
    
    public boolean valueState() {
      return VALUE_STATE.contains(state);
    }
    
    public boolean propertyState() {
      return PROPERTY_STATE.contains(state);
    }
    
    @Override
    public String toString() {
      return parent.toString();
    }
  }
  
  private static class NestedContext extends AbstractContext
  {
    ParserContext valueContext;
    
    NestedContext(ParserContext ctx) {
      super(ctx);
    }
    
    @Override
    public void init(ParserContext ctx) {
      this.valueContext = ctx;
    }
    
    @Override
    public JSONValue value() {
      if (valueContext != null) {
        return valueContext.value();
      }
      
      return JSONString.EMPTY;
    }
  }
  
  private static class PlainContext extends AbstractContext
  {
    final StringBuilder buf = new StringBuilder();
    
    PlainContext(ParserContext ctx) {
      super(ctx);
    }
    
    @Override
    public JSONValue value() {
      if (quoted()) {
        return new JSONString(buf.toString());
      }
      
      return handleNullValue(buf.toString());
    }
  }
  
  private static class HashContext extends AbstractContext
  {
    final JSONHash hash = new JSONHash();
    
    HashContext(ParserContext ctx) {
      super(ctx);
      
      state = State.PROPERTY;
    }
    
    @Override
    public JSONValue value() {
      return hash;
    }
    
    public void finishValue(String property, ParserContext root) {
      final JSONValue value = root.value();
      
      hash.put(property, value);
      state = State.AFTER_VALUE;
    }
  }
  
  private static class ArrayContext extends AbstractContext
  {
    final JSONArray array = new JSONArray();
    
    ArrayContext(ParserContext ctx) {
      super(ctx);
      
      state = State.VALUE;
    }
    
    @Override
    public JSONValue value() {
      return array;
    }
    
    public void finishValue(ParserContext root) {
      final JSONValue value = root.value();
      
      array.add(value);
      state = State.AFTER_VALUE;
    }
  }
  
  static final EnumSet<State> VALUE_STATE = EnumSet.of(State.VALUE, State.AFTER_VALUE, State.AFTER_PAIR); //NOSONAR
  static final EnumSet<State> PROPERTY_STATE = EnumSet.of(State.PROPERTY, State.AFTER_PROPERTY); //NOSONAR
  static final EnumSet<State> QUOTED = EnumSet.of(State.SINGLE_QUOTED_PROPERTY, State.SINGLE_QUOTED_VALUE, State.DOUBLE_QUOTED_PROPERTY, State.DOUBLE_QUOTED_VALUE); //NOSONAR

  private JSONUtil() {
    throw new AssertionError("Forbidden use of contructor!");
  }
}
