/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.client;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.EnumSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpEntity;
import org.apache.http.HttpStatus;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;

import cz.tdp.kshield.client.UserInfo.AuthType;

/**
 * <b>This is central class in client package.</b> This class encapsulates communication with KeyShield SSO Server.
 * 
 * @see cz.tdp.kshield.client.UserInfo
 * @see cz.tdp.kshield.integration.SimpleAuthenticationServiceImpl
 * @see cz.tdp.kshield.integration.web.DefaultAuthenticationServiceImpl
 */
public class KShieldClient
{
  private static final String JSON_USER_BY_IP_CMD = "/json/userByIP/";
  private static final String CLIENT_MESSAGE_CMD  = "/api/sendClientMessage";
  
  private final URI url;
  private final String apiKey;
  private final CloseableHttpClient httpclient;
  private final String attributes;
  private ResponseLevel responseLevel;
  
  /**
   * Construct a new KShieldClient for given KeyShield SSO Server url, given a HttpClient
   * 
   * @param url URL of KeyShield SSO Server web service (http://hostname:port).
   * 
   * @throws IllegalArgumentException if url is null, empty or invalid
   */
  public KShieldClient(String url) {
    this(createURI(url), null, null, null);
  }
  
  /**
   * Construct a new KShieldClient for given KeyShield SSO Server url, given a HttpClient
   * 
   * @param url URL of KeyShield SSO Server web service (http://hostname:port).
   * @param httpclient configured instance of apache HTTP client - for example configured for multithreaded usage, which allows sharing single instance of KShieldClient between multiple threads
   *                    Creates default http client if parameter is null.
   * @param attrs Additional attributes retrieved from KeyShield SSO Server
   * 
   * @throws IllegalArgumentException if url is null, empty or invalid
   */
  public KShieldClient(String url, CloseableHttpClient httpclient, String... attrs) {
    this(createURI(url), null, httpclient, createAttrs(attrs));
  }
  
  /**
   * Construct a new KShieldClient for given KeyShield SSO Server url, given a HttpClient
   * 
   * @param url URL of KeyShield SSO Server web service (http://hostname:port).
   * @param apiKey API authorization key.
   * @param httpclient configured instance of apache HTTP client - for example configured for multithreaded usage, which allows sharing single instance of KShieldClient between multiple threads
   *                    Creates default http client if parameter is null.
   * @param attrs Additional attributes retrieved from KeyShield SSO Server
   * 
   * @throws IllegalArgumentException if url is null, empty or invalid
   */
  public KShieldClient(String url, String apiKey, CloseableHttpClient httpclient, String... attrs) {
    this(createURI(url), apiKey, httpclient, createAttrs(attrs));
  }
  
  protected KShieldClient(URI url, String apiKey, CloseableHttpClient httpclient, String attrs) {
    if (httpclient == null) {
      this.httpclient = HttpClients.createDefault();
    }
    else {
      this.httpclient = httpclient;
    }
    
    this.url = url;
    this.apiKey = apiKey;
    this.attributes = attrs;
  }
  
  public ResponseLevel getResponseLevel() {
    return this.responseLevel;
  }

  public void setResponseLevel(ResponseLevel extendedResponse) {
    this.responseLevel = extendedResponse;
  }

  /**
   * Retrieves UserInfo for a given IP address
   * 
   * @param remoteIP IP address of workstation for which you want to obtain UserInfo
   * @return new UserInfo instance
   * @throws KShieldClientException
   */
  public UserInfo getUserByIP(String remoteIP) throws KShieldClientException {
    return getUserByIP(remoteIP, null);
  }
  
  /**
   * Retrieves UserInfo for a given IP address
   * 
   * @param remoteIP IP address of workstation for which you want to obtain UserInfo
   * @param usernameAttribute name of attribute for username, set it to null if you want use screenName instead of username
   * @return new UserInfo instance or null if authentication method is not supported
   * @throws KShieldClientException
   */
  public UserInfo getUserByIP(String remoteIP, String usernameAttribute) throws KShieldClientException {
    try {
      final URI uri = createUserByIPUri(remoteIP);
      
      return processGetUserByIP(uri, remoteIP, usernameAttribute);
    }
    catch (URISyntaxException e) {
      throw new KShieldClientException("Invalid KeyShield SSO Server request URL", e);
    }
    catch (IllegalStateException|IOException  e) {
      throw new KShieldClientException(e);
    }
  }

  private URI createUserByIPUri(String remoteIP) throws URISyntaxException {
    final URIBuilder builder = new URIBuilder(url)
        .setPath(url.getPath() + JSON_USER_BY_IP_CMD + remoteIP);
    
    if (apiKey != null && apiKey.length() > 0) {
      builder.addParameter("key", apiKey);
      
      if (responseLevel != null) {
        builder.addParameter("level", responseLevel.name());
      }
    }
    
    if (attributes != null && attributes.length() > 0) {
      builder.addParameter("attributes", attributes);
    }
    
    return builder.build();
  }

  private UserInfo processGetUserByIP(final URI uri, String remoteIP, String usernameAttribute) throws KShieldClientException, IOException {
    if (log.isDebugEnabled()) {
      log.debug("Querying KShield: " + uri);
    }
    
    final HttpGet httpget = new HttpGet(uri);

    try (CloseableHttpResponse response = httpclient.execute(httpget)) {
      final HttpEntity entity = response.getEntity();
 
      if (log.isDebugEnabled()) {
        log.debug("Response status: " + response.getStatusLine());
      }
      
      if (entity != null && response.getStatusLine() != null) {
        final int httpStatus = response.getStatusLine().getStatusCode();
        return createUserInfo(remoteIP, usernameAttribute, entity, httpStatus);
      }
      
      log.warn("Missing HttpResponse entity");
      throw new KShieldClientException("Missing HttpResponse entity");
    }
  }
  
  /**
   * @param from
   * @param to
   * @param message
   * @throws KShieldClientException
   */
  public void sendClientMessage(String from, String to, String message) throws KShieldClientException {
    sendClientMessage(new ClientMessage(from, to, message));
  }
  
  /**
   * Sends client message to KeyShield server
   * 
   * @param msg Message data
   * @throws KShieldClientException
   */
  public void sendClientMessage(ClientMessage msg) throws KShieldClientException {
    try {
      final URI uri = createClientMessageUri();
      
      processClientMessage(uri, msg);
    }
    catch (URISyntaxException e) {
      throw new KShieldClientException("Invalid KeyShield SSO Server request URL", e);
    }
    catch (IllegalStateException|IOException e) {
      throw new KShieldClientException(e);
    }
  }

  protected URI createClientMessageUri() throws URISyntaxException {
    final URIBuilder builder = new URIBuilder(url)
        .setPath(url.getPath() + CLIENT_MESSAGE_CMD);
    
    if (apiKey != null && apiKey.length() > 0) {
      builder.addParameter("key", apiKey);
      
      if (responseLevel != null) {
        builder.addParameter("level", responseLevel.name());
      }
    }
    
    return builder.build();
  }

  private void processClientMessage(URI uri, ClientMessage msg) throws IOException, KShieldClientException {
    if (log.isDebugEnabled()) {
      log.debug("Sending KShield command: " + uri);
    }
    
    final HttpPost httppost = new HttpPost(uri);
    httppost.setEntity(new StringEntity(msg.toJSON(), ContentType.APPLICATION_JSON));
    
    if (log.isDebugEnabled()) {
      log.debug("Client message: " + msg.toJSON());
    }
    
    // setting timeout to 10 sec and socket timeout to 2 minutes, we expect possible long ldap lookup on server side
    final RequestConfig config = RequestConfig.custom()
        .setSocketTimeout(120 * 1000)
        .setConnectTimeout(10 * 1000)
        .build();
    
    httppost.setConfig(config);
    
    try (CloseableHttpResponse response = httpclient.execute(httppost)) {
      final HttpEntity entity = response.getEntity();
  
      if (log.isDebugEnabled()) {
        log.debug("Response status: " + response.getStatusLine());
      }
      
      if (entity != null && response.getStatusLine() != null) {
        final int httpStatus = response.getStatusLine().getStatusCode();
        checkClientMessageResponse(entity, httpStatus);
      }
      else {
        log.warn("Missing HttpResponse entity");
        throw new KShieldClientException("Missing HttpResponse entity");
      }
    }
  }
  
  private void checkClientMessageResponse(HttpEntity entity, int httpStatus) throws KShieldClientException {
    if (httpStatus == HttpStatus.SC_FORBIDDEN) {
      throw new KShieldInvalidApiKeyException(apiKey());
    }
    if (httpStatus != HttpStatus.SC_OK) {
      throw new KShieldClientException("Invalid status returned: " + httpStatus);
    }
  }

  private String apiKey() {
    if (apiKey != null && apiKey.startsWith("key=")) {
      return apiKey.substring(4);
    }
    return "";
  }

  private UserInfo createUserInfo(String remoteIP, String usernameAttribute, HttpEntity entity, int httpStatus) throws KShieldClientException, IOException {
    final UserInfo userInfo = UserInfoJsonDecoder.decodeFromStream(entity.getContent());

    if (log.isDebugEnabled()) {
      log.debug("KShield response: " + userInfo.getUserID() + ", authType: " + userInfo.getAuthType() + ", ipAddress: " + userInfo.getIpAddress());
    }
    
    if (httpStatus == HttpStatus.SC_FORBIDDEN) {
      throw new KShieldInvalidApiKeyException(apiKey());
    }
    
    if (userInfo.getUserID() != null && allowedAuthTypes.contains(userInfo.getAuthType())) {
      if (!remoteIP.equals(userInfo.getIpAddress())) {
        throw new KShieldInvalidRemoteIpException(remoteIP, userInfo.getIpAddress());
      }
      
      userInfo.setUsernameAttribute(usernameAttribute);
      
      if (log.isDebugEnabled()) {
        log.debug("Authenticated username: " + userInfo.getUsername());
      }
      
      return userInfo;
    }
    
    return null;
  }
  
  /**
   * Retrieves UserInfo for a given IP address
   * @deprecated Please use instance methods instead
   * 
   * @param url KeyShield SSO Server url
   * @param remoteIP IP address of workstation for which you want to obtain UserInfo
   * @param usernameAttribute attribute for fetching username (could be null)
   * @return value of user id attribute fetched from KeyShield SSO Server
   */
  @Deprecated
  public static UserInfo getUserByIP(String url, String remoteIP, String usernameAttribute) {
    return getUserByIP(url, null, remoteIP, usernameAttribute);
  }

  /**
   * Retrieves UserInfo for a given IP address
   * @deprecated Please use instance methods instead
   * 
   * @param url KeyShield SSO Server url
   * @param remoteIP IP address of workstation for which you want to obtain UserInfo
   * @param usernameAttribute attribute for fetching username (could be null)
   * @return value of user id attribute fetched from KeyShield SSO Server
   */
  @Deprecated
  public static UserInfo getUserByIP(String url, String apiKey, String remoteIP, String usernameAttribute) {
    if (remoteIP != null && !remoteIP.isEmpty()) {
      try {
        final KShieldClient client;
        
        if (usernameAttribute != null) {
          client = new KShieldClient(url, apiKey, null, usernameAttribute);
        }
        else {
          client = new KShieldClient(url, apiKey, null);
        }
  
        return client.getUserByIP(remoteIP, usernameAttribute);
      }
      catch (KShieldClientException e) {
        log.error("client error", e);
      }
      catch (IllegalArgumentException e) {
        log.error("Invalid or empty KeyShield SSO Server URL", e);
      }
    }

    return null;
  }
  
  /**
   * Retrieves only username for a given IP address
   * @deprecated Please use instance methods instead
   * 
   * @param url KeyShield SSO Server url
   * @param remoteIP IP address of workstation for which you want to obtain username
   * @return username fetched from KeyShield SSO Server
   */
  @Deprecated
  public static String getUsernameByIP(String url, String remoteIP) {
    return getUsernameByIP(url, remoteIP, "cn");
  }
  
  /**
   * Retrieves only username for a given IP address
   * @deprecated Please use instance methods instead
   * 
   * @param url KeyShield SSO Server url
   * @param remoteIP IP address of workstation for which you want to obtain username
   * @param usernameAttribute name of username attribute, screenName is used if not defined
   * @return username fetched from KeyShield SSO Server
   */
  @Deprecated
  public static String getUsernameByIP(String url, String remoteIP, String usernameAttribute) {
    final UserInfo userInfo = getUserByIP(url, remoteIP, usernameAttribute);
    
    if (userInfo != null) {
      return userInfo.getUsername();
    }
    
    return null;
  }
  
  private EnumSet<AuthType> allowedAuthTypes = UserInfo.DEFAULT_ALLOWED_AUTH_TYPES;

  /**
   * Return allowed authentication types. See {@link AuthType}
   */
  public EnumSet<AuthType> getAllowedAuthTypes() {
    return allowedAuthTypes;
  }

  /**
   * Set allowed authentication types. See {@link AuthType}
   */
  public void setAllowedAuthTypes(EnumSet<AuthType> allowAuthTypes) {
    if (allowAuthTypes != null) {
      this.allowedAuthTypes = allowAuthTypes;
    }
    else {
      allowedAuthTypes = UserInfo.DEFAULT_ALLOWED_AUTH_TYPES;
    }
  }
  
  private static URI createURI(String url) {
    if (url == null || url.isEmpty()) throw new IllegalArgumentException("Missing or empty url parameter");
    return URI.create(url);
  }
  
  private static String createAttrs(String[] attrs) {
    final StringBuilder buf = new StringBuilder(attrs != null ? ((attrs.length + 1) * 10) : 30);
    
    buf.append(UserInfo.GUID_ATTRIBUTE);
    
    if (attrs != null && attrs.length > 0) {
      for (String attr : attrs) {
        if (!UserInfo.GUID_ATTRIBUTE.equalsIgnoreCase(attr)) {
          buf.append(',');
          buf.append(attr);
        }
      }
    }
    
    return buf.toString();
  }
  
  private static final Log log = LogFactory.getLog(KShieldClient.class);
}
