/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.client;

import java.io.Serializable;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * This class contains user data retrieved from the KeyShield SSO Server
 * @see cz.tdp.kshield.client.KShieldClient
 */
public class UserInfo implements Serializable
{
  private static final long serialVersionUID = 8935842552519353975L;
  
  static final String GUID_ATTRIBUTE = "x-guid";
  
  private String usernameAttribute;
  
  private String userID;
  private String connectorID;
  private String screenName;
  private String ipAddress;
  private AuthType authType;
  private String authMethod;
  private String client;
  private boolean hwTokenPresent;
  private boolean hwTokenEnabled;
  private boolean hwSwitching;
  
  private boolean manual;
  
  private final ClientInfo clientInfo = new ClientInfo();
  private Map<String,List<String>> attributes = Collections.emptyMap();
  
  /**
   * @return the authenticated user username or null if no user is authenticated.
   * If user ID attribute is defined, then return this attribute value, otherwise return screenName
   */
  public String getUsername() {
    if (usernameAttribute != null && usernameAttribute.length() > 0) {
      return usernameFromAttribute();
    }
    
    return screenName;
  }
  
  public String getGUID() {
    return guidFromAttribute();
  }
  
  /**
   * @return the screenname/username part from the user ID or null if no user is authenticated
   */
  public String getScreenName() {
    return screenName;
  }
  
  void setScreenName(String screenName) {
    this.screenName = screenName;
  }
  
  /**
   * @return KeyShield SSO Server user ID / FDN or null if no user is authenticated
   * For eDirectory it's LDAP FDN and for AD/domain it's domain://DOMAINNAME&#92username.
   */
  public String getUserID() {
    return userID;
  }
  
  
  void setUserID(String fdn) {
    this.userID = fdn;
  }
  
  /**
   * @return KeyShield SSO Server authentication type.
   * See {@link AuthType} for more info.
   */
  public AuthType getAuthType() {
    return authType;
  }
  
  void setAuthType(AuthType authType) {
    this.authType = authType;
  }
  
  
  /**
   * @return the SSO connector ID
   */
  public String getConnectorID() {
    return connectorID;
  }
  
  void setConnectorID(String connectorID) {
    this.connectorID = connectorID;
  }
  
  /**
   * @return the IP address of originating request
   */
  public String getIpAddress() {
    return ipAddress;
  }
  
  
  void setIpAddress(String ipAddress) {
    this.ipAddress = ipAddress;
  }
  
  
  /**
   * @return true if manual authentication method was used
   * @deprecated use authType instead
   */
  @Deprecated
  public boolean isManual() {
    return manual;
  }
  
  /**
   * @deprecated use authType instead
   */
  @Deprecated
  void setManual(boolean manual) {
    this.manual = manual;
  }
  
  /**
   * @return the actual used Authentication method
   */
  public String getAuthMethod() {
    return this.authMethod;
  }
  
  void setAuthMethod(String authMethod) {
    this.authMethod = authMethod;
  }
  
  /**
   * @return the client type or null
   */
  public String getClient() {
    return this.client;
  }
  
  void setClient(String client) {
    this.client = client;
  }
  
  /**
   * @return true if HW token is present
   */
  public boolean isHwTokenPresent() {
    return this.hwTokenPresent;
  }
  
  void setHwTokenPresent(boolean hwTokenPresent) {
    this.hwTokenPresent = hwTokenPresent;
  }
  
  /**
   * @return true if HW token is enabled
   */
  public boolean isHwTokenEnabled() {
    return this.hwTokenEnabled;
  }

  void setHwTokenEnabled(boolean hwTokenEnabled) {
    this.hwTokenEnabled = hwTokenEnabled;
  }

  /**
   * @return true if user switching is enabled
   */
  public boolean isHwSwitching() {
    return this.hwSwitching;
  }

  void setHwSwitching(boolean hwSwitching) {
    this.hwSwitching = hwSwitching;
  }

  /**
   * @return additional user attributes retrieved from KeyShield SSO Server
   * Map contains only attributes with at leas one value
   */
  public Map<String,List<String>> getAttributes() {
    return attributes;
  }
  
  /**
   * @param attributeName
   * @return non-null list of values for given attribute name
   */
  public List<String> getAttributeValues(String attributeName) {
    final List<String> values = attributes.get(attributeName.toLowerCase());
    
    if (values == null) {
      return Collections.emptyList();
    }
    return values;
  }
  
  /**
   * @param attributeName
   * @return first attribute value or null if value doesnt exists
   */
  public String getAttributeValue(String attributeName) {
    final List<String> values = attributes.get(attributeName.toLowerCase());
    
    if (values != null && !values.isEmpty()) {
      return values.get(0);
    }
    
    return null;
  }
  
  void setAttributes(Map<String,List<String>> attributes) {
    this.attributes = attributes;
  }
  
  public ClientInfo getClientInfo() {
    return this.clientInfo;
  }

  /**
   * @return name of username attribute
   */
  public String getUsernameAttribute() {
    return usernameAttribute;
  }
  
  /**
   * Set name of username attribute, if null screenName is used as username
   */
  public void setUsernameAttribute(String usernameAttribute) {
    this.usernameAttribute = usernameAttribute;
  }
  
  @Override
  public String toString() {
    return userID;
  }
  
  private String usernameFromAttribute() {
    final List<String> values = attributes.get(usernameAttribute.toLowerCase());
    
    if (values != null && !values.isEmpty()) {
      if (values.size() > 1) {
        log.warn("Multiple values for usernameAttribute: " + usernameAttribute.toLowerCase() + " = " + values);
      }
      
      return values.get(0);
    }
    
    log.warn("Cannot retrieve valid username attribute, returned attribute values: " + usernameAttribute.toLowerCase() + " = " + values);
    
    return null;
  }
  
  private String guidFromAttribute() {
    final List<String> values = attributes.get(GUID_ATTRIBUTE);
    
    if (values != null && !values.isEmpty()) {
      if (values.size() > 1) {
        log.warn("Multiple values for GUID: " + GUID_ATTRIBUTE + " = " + values);
      }
      
      return values.get(0);
    }
    
    log.warn("Cannot retrieve valid GUID attribute, returned attribute values: " + GUID_ATTRIBUTE + " = " + values);
    
    return null;
  }
  
  private static final Log log = LogFactory.getLog(UserInfo.class);
  
  public static final EnumSet<AuthType> DEFAULT_ALLOWED_AUTH_TYPES = EnumSet.complementOf(EnumSet.of(AuthType.UID, AuthType.NOT_AUTHENTICATED));
  
  /**
   * Supported authentication types returned by KeyShield SSO Server
   */
  public enum AuthType
  {
    /**
     * user is not authenticated
     */
    NOT_AUTHENTICATED(null),
    
    /**
     * user authenticated by UID.
     */
    UID("U"),
    
    /**
     * User authenticated using eDirectory method (User ID contains LDAP FDN e.g cn=user1,o=org).
     */
    EDIRECTORY("E"),
    
    /**
     * User authenticated using AD domain method (User ID contains domain://domainname&#92username ).
     */
    ACTIVEDIRECTORY("A"),
    
    /**
     * Only for backward compatibility
     * @deprecated Use type MANUAL instead
     */
    @Deprecated LDAP("L"),
    
    /**
     * Manual authentication method - not used in KeyShield SSO Server 1.3, since manual users are verified using LDAP
     * and User ID contains LDAP FDN. - use isManual()
     */
    MANUAL("M");
    
    private final String code;
    
    AuthType(String code) {
      this.code = code;
    }
    
    /**
     * @param code
     * @return AuthType based on code value, for unknown code returns NOT_AUTHENTICATED
     */
    public static AuthType fromCode(String code) {
      final AuthType val = CODE_TO_AUTH_TYPE_MAP.get(code);
      if (val != null) {
        return val;
      }
      
      return NOT_AUTHENTICATED;
    }
    
    /**
     * @return auth type code
     */
    public String code() {
      return code;
    }
    
    private static final Map<String,AuthType> CODE_TO_AUTH_TYPE_MAP = new HashMap<>();
    
    static {
      for (AuthType val : values()) {
        if (val.code != null) {
          CODE_TO_AUTH_TYPE_MAP.put(val.code, val);
        }
      }
    }
  }
  
  public static class ClientInfo
  {
    private String clientVersion;
    private String clientOS;
    private String clientOSVersion;
    private String clientHostname;
    private String stationKey;
    
    /**
     * @return kshield client version
     */
    public String getClientVersion() {
      return this.clientVersion;
    }
    
    void setClientVersion(String clientVersion) {
      this.clientVersion = clientVersion;
    }
    
    /**
     * @return client OS name
     */
    public String getClientOS() {
      return this.clientOS;
    }
    
    void setClientOS(String clientOS) {
      this.clientOS = clientOS;
    }
    
    /**
     * @return client OS version
     */
    public String getClientOSVersion() {
      return this.clientOSVersion;
    }
    
    void setClientOSVersion(String clientOSVersion) {
      this.clientOSVersion = clientOSVersion;
    }
    
    /**
     * @return client hostname
     */
    public String getClientHostname() {
      return this.clientHostname;
    }
    
    void setClientHostname(String clientHostname) {
      this.clientHostname = clientHostname;
    }
    
    /**
     * @return station key (usually same as hostname)
     */
    public String getStationKey() {
      if (this.stationKey != null) {
        return this.stationKey;
      }
      
      return this.clientHostname;
    }
    
    void setStationKey(String stationKey) {
      this.stationKey = stationKey;
    }
  }
}
