/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import cz.tdp.kshield.client.JSONUtil.JSONHash;
import cz.tdp.kshield.client.JSONUtil.JSONValue;
import cz.tdp.kshield.client.UserInfo.AuthType;
import cz.tdp.kshield.client.UserInfo.ClientInfo;

final class UserInfoJsonDecoder
{
  @SuppressWarnings("deprecation") // using attr manual - to be removed in the future
  static UserInfo decode(String json) throws KShieldParserException {
    try {
      final JSONValue value = JSONUtil.parse(json);
      
      if (value != null && value.isHash()) {
        final UserInfo userInfo = new UserInfo();
        final JSONHash userInfoHash = value.asHash();
        
        userInfo.setScreenName(userInfoHash.get("screenName").asString());
        userInfo.setUserID(userInfoHash.get("fdn").asString());
        userInfo.setIpAddress(userInfoHash.get("ipAddress").asString());
        userInfo.setManual(userInfoHash.get("manual").asBoolean());
        userInfo.setConnectorID(userInfoHash.get("connectorID").asString());
        userInfo.setClient(userInfoHash.get("client").asString());
        userInfo.setHwTokenPresent(userInfoHash.get("hwTokenPresent").asBoolean());
        userInfo.setAuthMethod(userInfoHash.get("authMethod").asString());
        userInfo.setAuthType(AuthType.fromCode(userInfoHash.get("authType").asString()));

        if (userInfoHash.get("clientInfo").isHash()) {
          final ClientInfo clientInfo = userInfo.getClientInfo();
          final JSONHash clientInfoHash = userInfoHash.get("clientInfo").asHash();
          
          clientInfo.setClientVersion(clientInfoHash.get("clientVersion").asString());
          clientInfo.setClientOS(clientInfoHash.get("clientOS").asString());
          clientInfo.setClientOSVersion(clientInfoHash.get("clientOSVersion").asString());
          clientInfo.setClientHostname(clientInfoHash.get("clientHostname").asString());
          clientInfo.setStationKey(clientInfoHash.get("stationKey").asString());
        }
            
        userInfo.setAttributes(toAttributes(value.asHash().get("attributes").asHash()));
        
        return userInfo;
      }
      else {
        log.error("Unexpected JSON response received: " + json);
        throw new KShieldParserException("Unexpected JSON response received");
      }
    }
    catch (KShieldParserException e) {
      // log received json string before rethrowing
      log.error("Invalid JSON received: " + json);
      throw e;
    }
  }
  
  static UserInfo decodeFromStream(InputStream content) throws IOException, KShieldParserException  {
    final InputStreamReader is = new InputStreamReader(content, "UTF-8");
    final StringBuilder sb = new StringBuilder(600);
    
    final char[] buffer = new char[200];
    
    int i = 0;
    while (i != -1 ) {
      i = is.read(buffer);
      if (i != -1) {
        sb.append(buffer, 0, i);
      }
    }

    return decode(sb.toString());
  }
  
  private static Map<String,List<String>> toAttributes(JSONHash hash) {
    final Map<String,List<String>> result = new HashMap<>();
    
    for (Map.Entry<String,JSONValue> entry : hash.hash().entrySet()) {
      result.put(entry.getKey().toLowerCase(), stringListValue(entry.getValue()));
    }
    
    return Collections.unmodifiableMap(result);
  }
  
  private static List<String> stringListValue(JSONValue value) {
    final List<String> result;
    
    if (value.isNull()) {
      result = Collections.emptyList();
    }
    else if (value.isArray()) {
      final List<String> strings = new ArrayList<>(value.asArray().size());
      
      for (JSONValue val : value.asArray().values()) {
        strings.add(val.asString());
      }
      
      result = Collections.unmodifiableList(strings);
    }
    else {
      result = Collections.singletonList(value.asString());
    }
    
    return result;
  }

  // hide utility class constructor
  private UserInfoJsonDecoder() {
    throw new AssertionError("Forbidden use of contructor!");
  }
  
  private static final Log log = LogFactory.getLog(UserInfoJsonDecoder.class);
}
