/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.integration;

import java.util.ArrayList;
import java.util.Collection;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * Holds information about current user session
 */
public final class KShieldContext
{
  private static final ThreadLocal<Boolean> kshield = new ThreadLocal<>();
  
  private static final Collection<Runnable> initSessionHooks = new ArrayList<>(1);
  private static final Collection<Runnable> startSessionHooks = new ArrayList<>(1);
  private static final Collection<Runnable> closeSessionHooks = new ArrayList<>(1);
  
  /**
   * Add task perfomed before any request to KeyShield SSO Server
   * 
   * @param hook
   */
  public static void addInitHook(Runnable hook) {
    initSessionHooks.add(hook);
  }
  
  /**
   * Add tak perfomed after succesful UserInfo retrieval and validation
   * 
   * @param hook
   */
  public static void addStartHook(Runnable hook) {
    startSessionHooks.add(hook);
  }
  
  /**
   * Add task performed when KeyShield SSO Server session is finished or in case of retrieval or validation failure
   * 
   * @param hook
   */
  public static void addCloseHook(Runnable hook) {
    closeSessionHooks.add(hook);
  }
  
  /**
   * Called before initiating KeyShield SSO Server session
   */
  public static void initKShieldSession() {
    kshield.remove();
    
    runHooks(initSessionHooks);
  }
  
  /**
   * Called after succesful UserInfo retrieval and validation
   */
  public static void startKShieldSession() {
    kshield.set(Boolean.TRUE);
    
    runHooks(startSessionHooks);
  }
  
  /**
   * Called when KeyShield SSO Server session is finished or in case of retrieval or validation failure
   */
  public static void closeKShieldSession() {
    kshield.remove();
    
    runHooks(closeSessionHooks);
  }
  
  /**
   * @return true if current user was succesfully identified to KeyShield SSO Server
   */
  public static boolean isKShieldSession() {
    final Boolean sess = kshield.get();
    return sess != null && sess;
  }
  
  private static void runHooks(Collection<Runnable> hooks) {
    for (Runnable hook : hooks) {
      try {
        hook.run();
      }
      catch (Exception e) {
        log.warn("Exception while executing kshield context hook", e);
      }
    }
  }
  
  private static final Log log = LogFactory.getLog(KShieldContext.class);
  
  private KShieldContext() {
    throw new UnsupportedOperationException();
  }
}
