/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.notification;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import cz.tdp.kshield.client.UserInfo;
import cz.tdp.kshield.integration.AuthenticationService;

public class KShieldAuthManagerImpl implements KShieldAuthManager
{
  private final AuthenticationService authService;
  
  private static final UserInfo NULL_AUTH = new UserInfo();
  private final ConcurrentMap<String,UserInfo> authenticatedUsers = new ConcurrentHashMap<>();
  
  private final ConcurrentMap<String,Collection<KShieldEndpoint>> endpoints = new ConcurrentHashMap<>();
  
  private final List<AuthChangeHandler> onLogin = new ArrayList<>(4);
  private final List<AuthChangeHandler> onLogout = new ArrayList<>(4);
  
  public KShieldAuthManagerImpl(AuthenticationService authService) {
    this.authService = authService;
  }
  
  @Override
  public void addAuth(UserInfo userInfo) {
    if (userInfo == null) throw new IllegalArgumentException("Missing userInfo parameter!");
    
    final String ipAddress = userInfo.getIpAddress();
    
    if (ipAddress != null) {
      if (log.isDebugEnabled()) {
        log.debug("add auth: ip=" + ipAddress + ", user=" + userInfo.getUserID());
      }
      
      authenticatedUsers.put(ipAddress, userInfo);
      handleOnLogin(userInfo);
      signal(ipAddress);
    }
    else {
      log.warn("add auth: missing ip address");
    }
  }

  @Override
  public void removeAuth(String ipAddress) {
    if (log.isDebugEnabled()) {
      log.debug("disable auth: ip=" + ipAddress);
    }
    
    final UserInfo userInfo = authenticatedUsers.put(ipAddress, NULL_AUTH);
    if (userInfo != null && userInfo.getIpAddress() != null) {
      handleOnLogout(userInfo);
    }
    signal(ipAddress);
  }
  
  @Override
  public void resetAuth(String ipAddress) {
    if (isNotEmpty(ipAddress)) {
      if (log.isDebugEnabled()) {
        log.debug("reset authentication for ip=" + ipAddress);
      }
      
      authenticatedUsers.remove(ipAddress);
      signal(ipAddress);
    }
    else {
      log.debug("reset all authentications");
      
      authenticatedUsers.clear();
      
      for (String addr : endpoints.keySet()) {
        signal(addr);
      }
    }
  }
  
  @Override
  public void registerOnLoginHandler(AuthChangeHandler handler) {
    onLogin.add(handler);
  }
  
  @Override
  public void registerOnLogoutHandler(AuthChangeHandler handler) {
    onLogout.add(handler);
  }
  
  private void signal(String ipAddress) {
    final Collection<KShieldEndpoint> list = endpoints.get(ipAddress);
    if (list != null) {
      signal(list);
    }
  }
  
  private void signal(Collection<KShieldEndpoint> list) {
    for (KShieldEndpoint ws : list) {
      if (ws.isActive()) {
        if (log.isDebugEnabled()) {
          log.debug("signalling WebSocket (" + ws + ")");
        }
        ws.signal();
      }
    }
  }

  @Override
  public boolean hasAuth(String ipAddress, String user) {
    UserInfo userInfo = authenticatedUsers.get(ipAddress);
    
    final String userID;
    
    if (userInfo == null) {
      userInfo = authService.createUserInfo(ipAddress);
      
      if (userInfo != null && userInfo.getUserID() != null) {
        userID = userInfo.getUserID();
        
        if (log.isDebugEnabled()) {
          log.debug("kshield auth changed: add ip=" + ipAddress + ", user=" + userID);
        }
        
        addAuth(userInfo);
      }
      else {
        if (log.isDebugEnabled()) {
          log.debug("kshield auth changed: remove ip=" + ipAddress);
        }
        
        removeAuth(ipAddress);
        
        userID = null;
      }
    }
    else {
      userID = userInfo.getUserID();
    }
    
    return user != null ? user.equals(userID) : isNotEmpty(userID);
  }

  private void handleOnLogin(UserInfo userInfo) {
    for (AuthChangeHandler handler : onLogin) {
      handler.handle(userInfo);
    }
  }
  
  private void handleOnLogout(UserInfo userInfo) {
    for (AuthChangeHandler handler : onLogout) {
      handler.handle(userInfo);
    }
  }
  
  @Override
  public void addEndpoint(String ipAddress, KShieldEndpoint ws) {
    if (log.isDebugEnabled()) {
      log.debug("adding WebSocket Endpoint ("+ws+") for ipAddress=" + ipAddress);
    }
        
    Collection<KShieldEndpoint> list = endpoints.get(ipAddress);
    if (list == null) {
      list = new CopyOnWriteArrayList<>();
      final Collection<KShieldEndpoint> origList = endpoints.putIfAbsent(ipAddress, list);
      if (origList != null) {
        list = origList;
      }
    }
    
    list.add(ws);
  }
  
  @Override
  public void removeEndpoint(String ipAddress, KShieldEndpoint ws) {
    if (ipAddress == null) throw new IllegalArgumentException("Missing ipAddress parameter!");
      
    if (log.isDebugEnabled()) {
      log.debug("removing WebSocket Endpoint ("+ws+") for ipAddress=" + ipAddress);
    }
    
    final Collection<KShieldEndpoint> list = endpoints.get(ipAddress);
    if (list != null) {
      list.remove(ws);
    }
  }
  
  @Override
  public String poll(String ipAddress, String userID, boolean authenticated) {
    if (log.isDebugEnabled()) {
      log.debug("poll: " + ipAddress + ", userID="+userID + ", authenticated=" + authenticated);
    }
    
    final String result;
    
    if (authenticated) {
      if (isNotEmpty(userID)) {
        if (hasAuth(ipAddress, userID)) {
          result = "LOGIN";
        }
        else {
          result = "LOGOUT";
        }
      }
      else {
        result = "AUTH";
      }
    }
    else if (hasAuth(ipAddress, null)) {
      result = "LOGIN";
    }
    else {
      result = "OFF";
    }
    
    return result;
  }
  
  private boolean isNotEmpty(String userID) {
    return userID != null && !userID.isEmpty();
  }

  private static final Log log = LogFactory.getLog(KShieldAuthManagerImpl.class);
}
